==> spread-5.5.0.34-py3-none-any.whl: spread_knx/commands.py <==
"""
Взято из https://github.com/thelsing/knx/blob/master/src/knx/cemi_frame.cpp

cEMI Frame Format
+---------+--------+--------+--------+--------+---------+---------+--------+---------+
    | Header  |  Msg   |Add.Info| Ctrl 1 | Ctrl 2 | Source  | Dest.   |  Data  |   APDU  |
    |         | Code   | Length |        |        | Address | Address | Length |         |
    +---------+--------+--------+--------+--------+---------+---------+--------+---------+
      6 bytes   1 byte   1 byte   1 byte   1 byte   2 bytes   2 bytes   1 byte   2 bytes

        Header          = See below the structure of a cEMI header
        Message Code    = See below. On Appendix A is the list of all existing EMI and cEMI codes
        Add.Info Length = 0x00 - no additional info
        Control Field 1 =
        Control Field 2 =
        Source Address  = 0x0000 - filled in by router/gateway with its source address which is
                          part of the KNX subnet
        Dest. Address   = KNX group or individual address (2 byte)
        Data Length     = Number of bytes of data in the APDU excluding the TPCI/APCI bits
        APDU            = Application Protocol Data Unit - the actual payload including transport
                          protocol control information (TPCI), application protocol control
                          information (APCI) and data passed as an argument from higher layers of
                          the KNX communication stack

Control Field 1

          Bit  |
         ------+---------------------------------------------------------------
           7   | Frame Type  - 0x0 for extended frame
               |               0x1 for standard frame
         ------+---------------------------------------------------------------
           6   | Reserved
               |
         ------+---------------------------------------------------------------
           5   | Repeat Flag - 0x0 repeat frame on medium in case of an error
               |               0x1 do not repeat
         ------+---------------------------------------------------------------
           4   | System Broadcast - 0x0 system broadcast
               |                    0x1 broadcast
         ------+---------------------------------------------------------------
           3   | Priority    - 0x0 system
               |               0x1 normal
         ------+               0x2 urgent
           2   |               0x3 low
               |
         ------+---------------------------------------------------------------
           1   | Acknowledge Request - 0x0 no ACK requested
               | (L_Data.req)          0x1 ACK requested
         ------+---------------------------------------------------------------
           0   | Confirm      - 0x0 no error
               | (L_Data.con) - 0x1 error
         ------+---------------------------------------------------------------

 Control Field 2

          Bit  |
         ------+---------------------------------------------------------------
           7   | Destination Address Type - 0x0 individual address
               |                          - 0x1 group address
         ------+---------------------------------------------------------------
          6-4  | Hop Count (0-7)
         ------+---------------------------------------------------------------
          3-0  | Extended Frame Format - 0x0 standard frame
         ------+---------------------------------------------------------------
"""
from . address import Address


class DPTU:
    CMD_READ = 0x0
    CMD_STATE = 0x1
    CMD_SET = 0x2
    CMD_WRITE = 0xA
    CMD_VERBOSE = ['get', 'state', 'set', 'write']

    def __init__(self, command: int, data: bytes = b'', empty=False):
        self.command = command
        self.data = bytes(data)
        self.is_empty = empty

    def __str__(self):
        return '%s %s %s' % (self.CMD_VERBOSE[self.command], self.__class__.__name__, self.to_string)

    def fill(self, frame):
        self.is_empty = False
        self.data = bytearray([frame[1] & 0x3F]) + frame[2:]

    @classmethod
    def of(cls, frame: bytes):
        return cls((int.from_bytes(frame[:2], 'big') >> 6) & 0xF, bytearray([frame[1] & 0x3F]) + frame[2:])

    @classmethod
    def empty(cls):
        return cls(cls.CMD_STATE, empty=True)

    @property
    def length(self): return len(self.data)

    @property
    def frame(self) -> bytes:
        frame = self.command << 6
        if self.length > 0:
            frame += self.data[0]

        frame = frame.to_bytes(2, 'big') + self.data[1:]

        # if len(frame) == 1:
        #     frame = b'\x00' + frame

        return frame

    @property
    def is_request(self): return self.command == self.CMD_READ

    @property
    def value(self):
        return self.data.hex().upper()

    @property
    def to_string(self):
        value = self.value
        if value is None:
            return ''

        return value


class Switching(DPTU):
    @classmethod
    def set(cls, address: Address, value: bool):
        return KnxFrame(address, dptu=cls(cls.CMD_SET, value.to_bytes(1, 'big')))

    @classmethod
    def get(cls, address):
        return KnxFrame(address, dptu=cls(cls.CMD_READ, b'\x00'))

    @property
    def value(self):
        if not self.is_request and len(self.data) > 0:
            if not self.is_empty:
                return self.data[0] & 0x01 == 0x01

        return None


class Level(DPTU):
    @classmethod
    def set(cls, address, value: int):
        assert value in range(0, 256)
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_SET, value.to_bytes(2, 'big')))

    @classmethod
    def get(cls, address: Address):
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_READ, b'\x00\x00'))

    @property
    def value(self):
        if not self.is_request:
            if not self.is_empty:
                return int.from_bytes(self.data, 'big')

        return None


class Step(DPTU):
    @classmethod
    def set(cls, address: Address, up=True):
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_SET, int(up).to_bytes(1, 'big')))

    @property
    def value(self):
        if not self.is_request:
            if not self.is_empty:
                val = self.data[0]
                if val == 1:
                    return "Up"
                elif val == 0:
                    return "Down"
                return 'X3'

        return None


class KnxFrame:
    PRIORITY_SYSTEM = 0
    PRIORITY_NORMAL = 1
    PRIORITY_HIGH = 2
    PRIORITY_LOW = 3

    __slots__ = ['standard_frame', 'repeat', 'is_broadcast', 'priority', 'ack', 'error', 'is_group_address',
                 'address',
                 'routing_counter', 'length',
                 'dptu']

    def __init__(self, address, dptu: DPTU,
                 standard_frame=True,
                 repeat=False,
                 is_broadcast=True,
                 priority=PRIORITY_LOW,
                 ack=False,
                 error=False,

                 is_group_address=True,
                 routing_counter=6,
                 ):

        self.standard_frame = standard_frame
        self.repeat = repeat
        self.is_broadcast = is_broadcast
        self.priority = priority
        self.ack = ack
        self.error = error

        self.address = address

        self.is_group_address = is_group_address
        self.routing_counter = routing_counter
        self.length = dptu.length
        self.dptu = dptu

    def __str__(self):
        return '%s %s' % (self.address, self.dptu)

    @property
    def b0(self):
        return sum((
            self.standard_frame << 7,
            (not self.repeat) << 5,
            self.is_broadcast << 4,
            self.priority << 2,
            self.ack << 1,
            self.error
        ))

    @property
    def b2(self):
        addr_indexes = [11, 8, 0]
        return sum([int(i) << addr_indexes.pop(0) for i in self.address.split('/')])

    @property
    def drl(self):
        return (self.is_group_address << 7) + (self.routing_counter << 4) + self.dptu.length

    @property
    def frame(self):
        return bytearray([self.b0, 0, 0, *self.b2.to_bytes(2, 'big'), self.drl, *self.dptu.frame])

    @property
    def frame_str(self):
        return ''.join([hex(b)[2:].rjust(2, '0').upper() for b in self.frame])

    @classmethod
    def of(cls, frame: (bytes, str)):
        if isinstance(frame, str):
            frame = bytes.fromhex(frame)

        if len(frame) < 6:
            raise
            # raise FrameError()
        return cls(
            #  +---+---+---+---+-------+---+---+
            #  | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
            #  +---+---+---+---+-------+---+---+
            #  | T |...| R | B | p1 p0 | A | E |
            #  +---+---+---+---+-------+---+---+
            standard_frame=frame[0] >> 7 == 1,
            repeat=frame[0] & 0x20 == 0x0,
            is_broadcast=frame[0] & 0x10 == 0x10,
            priority=(frame[0] >> 2) & 0x3,
            ack=frame[0] & 0x2 == 0x2,
            error=frame[0] & 0x1 == 0x1,

            #  +----+----+----+----+----+----+---+----+---+---+---+---+---+---+---+-+--+
            #  | 15 | 14 | 13 | 12 | 11 | 10 | 9 | 8  | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0  |
            #  +----+-------------------+-------------+--------------------------------+
            #  | ...| H3   H2   H1   H0 | М2   M1  M0 | U7  U6  U5  U4  U3  U2  U1  U0 |
            #  +----+-------------------+-------------+---+---+---+---+---+---+---+-+--+
            address='/'.join([str(i) for i in [frame[3] >> 3 & 0xF, frame[3] & 0x7, frame[4]]]),

            #  +---+---+---+----+-------+---+----+
            #  | 7 | 6 | 5 | 4  | 3 | 2 | 1 | 0  |
            #  +---+---+---+----+----------------+
            #  | D | R2  R1  R0 | L3  L2  L1  L0 |
            #  +---+---+--------+----------------+

            is_group_address=frame[5] >> 7 == 1,
            routing_counter=(frame[5] >> 4) & 0x7,
            # length=frame[5] & 0xF,
            dptu=DPTU.of(frame[6:])
        )


# f0 = 'BC 00 00 01 01 E1 00 81'
# f0 = 'BC 11 08 00 02 E1 00 81'
# print(f0)
# f = KnxFrame.of(bytes.fromhex(f0))
# print(f.frame.hex().upper())

# f = Dimming.get(Address(control='4/0/2', state='4/0/26'))
# print(f.frame.hex().upper())

==> spread-5.8.1.80-py3-none-any.whl: spread_knx/commands.py <==
"""
Взято из https://github.com/thelsing/knx/blob/master/src/knx/cemi_frame.cpp

cEMI Frame Format
+---------+--------+--------+--------+--------+---------+---------+--------+---------+
    | Header  |  Msg   |Add.Info| Ctrl 1 | Ctrl 2 | Source  | Dest.   |  Data  |   APDU  |
    |         | Code   | Length |        |        | Address | Address | Length |         |
    +---------+--------+--------+--------+--------+---------+---------+--------+---------+
      6 bytes   1 byte   1 byte   1 byte   1 byte   2 bytes   2 bytes   1 byte   2 bytes

        Header          = See below the structure of a cEMI header
        Message Code    = See below. On Appendix A is the list of all existing EMI and cEMI codes
        Add.Info Length = 0x00 - no additional info
        Control Field 1 =
        Control Field 2 =
        Source Address  = 0x0000 - filled in by router/gateway with its source address which is
                          part of the KNX subnet
        Dest. Address   = KNX group or individual address (2 byte)
        Data Length     = Number of bytes of data in the APDU excluding the TPCI/APCI bits
        APDU            = Application Protocol Data Unit - the actual payload including transport
                          protocol control information (TPCI), application protocol control
                          information (APCI) and data passed as an argument from higher layers of
                          the KNX communication stack

Control Field 1

          Bit  |
         ------+---------------------------------------------------------------
           7   | Frame Type  - 0x0 for extended frame
               |               0x1 for standard frame
         ------+---------------------------------------------------------------
           6   | Reserved
               |
         ------+---------------------------------------------------------------
           5   | Repeat Flag - 0x0 repeat frame on medium in case of an error
               |               0x1 do not repeat
         ------+---------------------------------------------------------------
           4   | System Broadcast - 0x0 system broadcast
               |                    0x1 broadcast
         ------+---------------------------------------------------------------
           3   | Priority    - 0x0 system
               |               0x1 normal
         ------+               0x2 urgent
           2   |               0x3 low
               |
         ------+---------------------------------------------------------------
           1   | Acknowledge Request - 0x0 no ACK requested
               | (L_Data.req)          0x1 ACK requested
         ------+---------------------------------------------------------------
           0   | Confirm      - 0x0 no error
               | (L_Data.con) - 0x1 error
         ------+---------------------------------------------------------------

 Control Field 2

          Bit  |
         ------+---------------------------------------------------------------
           7   | Destination Address Type - 0x0 individual address
               |                          - 0x1 group address
         ------+---------------------------------------------------------------
          6-4  | Hop Count (0-7)
         ------+---------------------------------------------------------------
          3-0  | Extended Frame Format - 0x0 standard frame
         ------+---------------------------------------------------------------
"""
from . address import Address


class DPTU:
    CMD_READ = 0x0
    CMD_STATE = 0x1
    CMD_SET = 0x2
    CMD_WRITE = 0xA
    CMD_VERBOSE = ['get', 'state', 'set', 'write']

    def __init__(self, command: int, data: bytes = b'', empty=False):
        self.command = command
        self.data = bytes(data)
        self.is_empty = empty

    def __str__(self):
        return '%s %s %s' % (self.CMD_VERBOSE[self.command], self.__class__.__name__, self.to_string)

    def fill(self, frame):
        self.is_empty = False
        self.data = bytearray([frame[1] & 0x3F]) + frame[2:]

    @classmethod
    def of(cls, frame: bytes):
        return cls((int.from_bytes(frame[:2], 'big') >> 6) & 0xF, bytearray([frame[1] & 0x3F]) + frame[2:])

    @classmethod
    def empty(cls):
        return cls(cls.CMD_STATE, empty=True)

    @property
    def length(self): return len(self.data)

    @property
    def frame(self) -> bytes:
        frame = self.command << 6
        if self.length > 0:
            frame += self.data[0]

        frame = frame.to_bytes(2, 'big') + self.data[1:]

        # if len(frame) == 1:
        #     frame = b'\x00' + frame

        return frame

    @property
    def is_request(self): return self.command == self.CMD_READ

    @property
    def value(self):
        return self.data.hex().upper()

    @property
    def to_string(self):
        value = self.value
        if value is None:
            return ''

        return value


class Switching(DPTU):
    @classmethod
    def set(cls, address, value: bool):
        return KnxFrame(address, dptu=cls(cls.CMD_SET, value.to_bytes(1, 'big')))

    @classmethod
    def get(cls, address):
        return KnxFrame(address, dptu=cls(cls.CMD_READ, b'\x00'))

    @property
    def value(self):
        if not self.is_request and len(self.data) > 0:
            if not self.is_empty:
                return self.data[0] & 0x01 == 0x01

        return None


class Level(DPTU):
    @classmethod
    def set(cls, address, value: int):
        assert value in range(0, 256)
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_SET, value.to_bytes(2, 'big')))

    @classmethod
    def get(cls, address):
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_READ, b'\x00'))

    @property
    def value(self):
        if not self.is_request:
            if not self.is_empty:
                return int.from_bytes(self.data, 'big')

        return None


class Step(DPTU):
    @classmethod
    def set(cls, address, up=True):
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_SET, int(up).to_bytes(1, 'big')))

    @property
    def value(self):
        if not self.is_request:
            if not self.is_empty:
                val = self.data[0]
                if val == 1:
                    return "Up"
                elif val == 0:
                    return "Down"
                return 'X3'

        return None


class Float(DPTU):
    """ EIS5 """

    @classmethod
    def set(cls, address, value: float):
        raise NotImplemented()

    @classmethod
    def get(cls, address: Address):
        assert address is not None, "Адрес для выполнения команды не указан!!!"
        return KnxFrame(address, dptu=cls(cls.CMD_READ, b'\x00'))

    @property
    def value(self):
        if not self.is_request:
            if not self.is_empty:
                if self.data:
                    d = int(self.data[-2:].hex(), 16)
                    s = (d & 0b1000000000000000) >> 15
                    e = (d & 0b111100000000000) >> 11
                    m = d & 0b11111111111
                    if s == 1:
                        m -= 2048
                    res = (m << e) / 100
                    return res

        return None


class KnxFrame:
    PRIORITY_SYSTEM = 0
    PRIORITY_NORMAL = 1
    PRIORITY_HIGH = 2
    PRIORITY_LOW = 3

    __slots__ = ['standard_frame', 'repeat', 'is_broadcast', 'priority', 'ack', 'error', 'is_group_address',
                 'address',
                 'routing_counter', 'length',
                 'dptu']

    def __init__(self, address, dptu: DPTU,
                 standard_frame=True,
                 repeat=False,
                 is_broadcast=True,
                 priority=PRIORITY_LOW,
                 ack=False,
                 error=False,

                 is_group_address=True,
                 routing_counter=6,
                 ):

        self.standard_frame = standard_frame
        self.repeat = repeat
        self.is_broadcast = is_broadcast
        self.priority = priority
        self.ack = ack
        self.error = error

        self.address = address

        self.is_group_address = is_group_address
        self.routing_counter = routing_counter
        self.length = dptu.length
        self.dptu = dptu

    def __str__(self):
        return '%s %s' % (self.address, self.dptu)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.frame == other.frame

        return False

    @property
    def b0(self):
        return sum((
            self.standard_frame << 7,
            (not self.repeat) << 5,
            self.is_broadcast << 4,
            self.priority << 2,
            self.ack << 1,
            self.error
        ))

    @property
    def b2(self):
        addr_indexes = [11, 8, 0]
        return sum([int(i) << addr_indexes.pop(0) for i in self.address.split('/')])

    @property
    def drl(self):
        return (self.is_group_address << 7) + (self.routing_counter << 4) + self.dptu.length

    @property
    def frame(self):
        return bytearray([self.b0, 0, 0, *self.b2.to_bytes(2, 'big'), self.drl, *self.dptu.frame])

    @property
    def frame_str(self):
        return self.frame.hex().upper()

    @classmethod
    def of(cls, frame: (bytes, str)):
        if isinstance(frame, str):
            frame = bytes.fromhex(frame)

        if len(frame) < 6:
            raise
            # raise FrameError()
        return cls(
            #  +---+---+---+---+-------+---+---+
            #  | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
            #  +---+---+---+---+-------+---+---+
            #  | T |...| R | B | p1 p0 | A | E |
            #  +---+---+---+---+-------+---+---+
            standard_frame=frame[0] >> 7 == 1,
            repeat=frame[0] & 0x20 == 0x0,
            is_broadcast=frame[0] & 0x10 == 0x10,
            priority=(frame[0] >> 2) & 0x3,
            ack=frame[0] & 0x2 == 0x2,
            error=frame[0] & 0x1 == 0x1,

            #  +----+----+----+----+----+----+---+----+---+---+---+---+---+---+---+-+--+
            #  | 15 | 14 | 13 | 12 | 11 | 10 | 9 | 8  | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0  |
            #  +----+-------------------+-------------+--------------------------------+
            #  | ...| H3   H2   H1   H0 | М2   M1  M0 | U7  U6  U5  U4  U3  U2  U1  U0 |
            #  +----+-------------------+-------------+---+---+---+---+---+---+---+-+--+
            address='/'.join([str(i) for i in [frame[3] >> 3 & 0xF, frame[3] & 0x7, frame[4]]]),

            #  +---+---+---+----+-------+---+----+
            #  | 7 | 6 | 5 | 4  | 3 | 2 | 1 | 0  |
            #  +---+---+---+----+----------------+
            #  | D | R2  R1  R0 | L3  L2  L1  L0 |
            #  +---+---+--------+----------------+

            is_group_address=frame[5] >> 7 == 1,
            routing_counter=(frame[5] >> 4) & 0x7,
            # length=frame[5] & 0xF,
            dptu=DPTU.of(frame[6:])
        )


# f0 = 'BC 00 00 01 01 E1 00 81'
# f0 = 'BC 11 08 00 02 E1 00 81'
# print(f0)
# f = KnxFrame.of(bytes.fromhex(f0))
# print(f.frame.hex().upper())

# f = Dimming.get(Address(control='4/0/2', state='4/0/26'))
# print(f.frame.hex().upper())
